// -*- C++ -*-
/*!
 * @file RtcSubscriber.cpp
 * @brief Subscriber active object class runs on Output port
 * @date $Date$
 *
 * Copyright (c) 2004 Noriaki Ando <n-ando@aist.go.jp>
 *          National Institute of Industrial Science (AIST), Japan. All rights reserved.
 *
 * $Id$
 */

//#include <ace/ACE.h>
#include "RtcSubscriber.h"

namespace RTM
{

  using namespace std;

  //------------------------------------------------------------
  // SubscriberBase
  //------------------------------------------------------------
  SubscriberBase::SubscriberBase(OutPortBase* parent, InPort_ptr inport,
				   const SubscriptionID id, SubscriberProfile profile)
	: m_pParent(parent), m_pInPort(InPort::_duplicate(inport)),
	  m_Id(id), m_Profile(profile)
	  {
//		RTC_TRACE("SubscriberBase::SubscriberBase()");
//		open(0);
	  };

  SubscriberBase::~SubscriberBase()
  {
	//RTC_TRACE("SubscriberBase::~SubscriberBase()");

	this->release();
  }

  int SubscriberBase::open(void *args)
  {
	//RTC_TRACE("SubscriberBase::open(void *args)");
	
	m_Running = true;
	activate();
	return 0;
  }

  int SubscriberBase::close(unsigned long flags)
  {
	CORBA::release(m_pInPort);
	delete this;
	return 0;
  }

  void SubscriberBase::release()
  {
	//RTC_TRACE("SubscriberBase::release()");

	if (is_running())
	  {
		// The thread will be restarted by cond signal and completely stoped.
		this->stop_thread();
	  }
  }

  SubscriberProfile SubscriberBase::get_profile()
  {
	//RTC_TRACE("SubscriberBase::get_profile()");
	return m_Profile;
  }


  InPort_ptr SubscriberBase::get_inport()
  {
	//RTC_TRACE("SubscriberBase::get_inport()");
	return m_pInPort;
  }

  std::string SubscriberBase::get_id()
  {
	//RTC_TRACE("SubscriberBase::get_id()");
	return m_Id;
  }
  
  bool SubscriberBase::is_running()
  {
	//RTC_TRACE("SubscriberBase::is_running()");

	return m_Running;
  }

  void SubscriberBase::stop_thread()
  {
	//RTC_TRACE("SubscriberBase::stop_thread()");
	m_Running = false;
//	update(); // through cond. var. wait. is svc().
//	wait();   // wait finished svc() thread.
  }

  //------------------------------------------------------------
  // SubscriberOnce
  //------------------------------------------------------------
  SubscriberOnce::SubscriberOnce(OutPortBase* parent, InPort_ptr inport,
								 const SubscriptionID id,
								 SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	//RTC_TRACE("SubscriberOnce::SubscriberOnce()");
	open(0);
  }

  // Notify data update to subscriber
  bool SubscriberOnce::update()
  {
	//RTC_TRACE("SubscriberOnce::update()");
	if (m_NewData.m_Mutex.tryacquire_write() != 0)
	  {
		return false;
	  }

	m_NewData.m_Updated = true;
	m_NewData.m_Cond.signal();

	m_NewData.m_Mutex.release();
	return true;
  }

  // Subscriber active object main routine
  int SubscriberOnce::svc(void)
  {
	//RTC_TRACE("SubscribetrOnce::svc()");

	m_NewData.m_Mutex.acquire();
	
	// Waiting for new data updated
	while (!m_NewData.m_Updated)
	  m_NewData.m_Cond.wait();
	
	if (m_NewData.m_Updated)
	  {
		//		m_pParent->push(m_InPort);
		//		m_pParent->unsubscribe(m_InPort, m_Id);
		m_NewData.m_Updated = false;
	  }
	
	m_NewData.m_Mutex.release();
  }
  


  //------------------------------------------------------------
  // SubscriberPeriodic
  //------------------------------------------------------------
  SubscriberPeriodic::SubscriberPeriodic(OutPortBase* parent,
										 InPort_ptr inport,
										 const SubscriptionID id,
										 SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	//RTC_TRACE("SubscriberPeriodic::SubscriberPeriodic()");
	open(0);
  }


  // Notify data update to subscriber
  bool SubscriberPeriodic::update()
  {
	//RTC_TRACE("SubscriberPeriodic::update()");

	// Do nothing: Periodic subscriber object push data asynchronously to InPort
	return true;
  }


  // Subscriber active object main routine
  int SubscriberPeriodic::svc(void)
  {
	//RTC_TRACE("SubscriberPeriodic::svc()");
#ifdef ARTLinux
	//	art_enter();
#endif // ARTLinux
	while (m_Running)
	  {
		sleep(1);
		//		m_pParent->push(m_InPort, m_Id);
		
		// sleep() is needed
#ifdef ARTLinux
		//art_wait();
#else
		//		ACE_OS::sleep(m_WaitTime);
#endif //ARTLinux
		
	  }
  }


  //------------------------------------------------------------
  // SubscriberNew
  //------------------------------------------------------------
  SubscriberNew::SubscriberNew(OutPortBase* parent, InPort_ptr inport,
								const SubscriptionID id,
							   SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	//RTC_TRACE("SubscriberNew::SubscriberNew()");
	open(0);
  }


  // Notify data update to subscriber
  bool SubscriberNew::update()
  {
	//RTC_TRACE("SubscriberNew::update()");
	if (m_NewData.m_Mutex.tryacquire() != 0)
	  {
		return false;
	  }

	m_NewData.m_Updated = true;
	m_NewData.m_Cond.signal();

	m_NewData.m_Mutex.release();
	return true;
  }


  int SubscriberNew::svc(void)
  {
	//RTC_TRACE("SubscriberNew::svc()");
	
	while (m_Running)
	  {
		//		ACE_Guard<ACE_Thread_Mutex> guard(m_NewData.m_Mutex);
		m_NewData.m_Mutex.acquire();
		
		// Waiting for new data updated
		while (!m_NewData.m_Updated && m_Running)
		  {
			m_NewData.m_Cond.wait();
		  }
		
		if (m_NewData.m_Updated)
		  {
			if (m_pParent->push(m_pInPort, m_Id) != RTM_OK) 
			  {
				;
			  }
			m_NewData.m_Updated = false;
		  }
		m_NewData.m_Mutex.release();	
	  }

	// If subscriber comes here, subscriber will be finished and
	// be destructed immediately.

	return 0;
  }


  //------------------------------------------------------------
  // SubscriberTriggerd
  //------------------------------------------------------------
  SubscriberTriggerd::SubscriberTriggerd(OutPortBase* parent,
										 InPort_ptr inport,
										 const SubscriptionID id,
										 SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	//RTC_TRACE("SubscriberTriggerd::SubscriberTriggerd()");
	open(0);

  }


  //------------------------------------------------------------
  // SubscriberNewPeriodic
  //------------------------------------------------------------
  SubscriberNewPeriodic::SubscriberNewPeriodic(OutPortBase* parent,
											   InPort_ptr inport,
											   const SubscriptionID id,
											   SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	//RTC_TRACE("SubscriberNewPeriodic::SubscriberNewPeriodic()");
  }


  //------------------------------------------------------------
  // SubscriberPeriodicNew
  //------------------------------------------------------------
  SubscriberPeriodicNew::SubscriberPeriodicNew(OutPortBase* parent,
											   InPort_ptr inport,
											   const SubscriptionID id,
											   SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	//RTC_TRACE("SubscriberPeriodicNew::SubscriberPeriodicNew()");
	open(0);

  }



  //------------------------------------------------------------
  // SubscriberPeriodicTriggerd
  //------------------------------------------------------------
  SubscriberPeriodicTriggerd::
  SubscriberPeriodicTriggerd(OutPortBase* parent,
							 InPort_ptr inport,
							 const SubscriptionID id,
							 SubscriberProfile profile)
	: SubscriberBase(parent, inport, id, profile)
  {
	//RTC_TRACE("SubscriberPeriodicTriggerd::SubscriberPeriodicTriggerd()");
	open(0);

  }



  //------------------------------------------------------------
  // SubscriberTriggerdPeriodic
  //------------------------------------------------------------

  SubscriberTriggerdPeriodic::
  SubscriberTriggerdPeriodic(OutPortBase* parent,
							 InPort_ptr inport,
							 const SubscriptionID id,
							 SubscriberProfile profile)
	:SubscriberBase(parent, inport, id, profile)
  {
	//RTC_TRACE("SubscriberTriggerdPeriodic::SubscriberTriggerdPeriodic()");
	open(0);
  }




}; //end of namespace RTM






