// -*- C++ -*-
/*!
 * @file RtcSubscriber.h
 * @brief Subscriber active object class runs on Output port
 * @date $Date$
 *
 * Copyright (c) 2004 Noriaki Ando <n-ando@aist.go.jp>
 *          National Institute of Industrial Science (AIST), Japan. All rights reserved.
 *
 * $Id$
 */

#ifndef __RtcSubscriber_h__
#define __RtcSubscriber_h__

#include "rtm/RtcOutPortBase.h"
#include "rtm/RtcInPortBase.h"

#include "rtm/Rtc_Mutex.h"
#include "rtm/Rtc_Task.h"

namespace RTM {
  //============================================================
  /*!
   * @class SubscriberBase
   * 
   * @brief Subscriber object.
   * This is derived from ACE_Task active object and 
   * observer pattern is alos applied.
   */ 
  /*!
   * @if jp
   *
   * @class SubscriberBase
   *
   * @brief Subscriber 쥯饹
   * 
   * Subscriber* δ쥯饹 SubscriberBase.
   * Ƽ Subscriber ϤΥ饹Ѿƾܺ٤롣
   *
   * @else
   *
   * @class SubscriberBase
   *
   * @brief Base class of subscriber.
   *
   * A base class of Subscriber*.
   * Variation of Subscriber* which implements details of subscriber
   * inherits this SubscriberBase class.
   *
   * @endif
   */
  class SubscriberBase
	: public RTC_Task
//	: public ACE_Task<ACE_MT_SYNCH>
  {
  public:
	/*!
	 * @if jp
	 *
	 * @brief SubscriberBase 饹󥹥ȥ饯
	 *
	 * @param parent Subscriber* 򥤥󥹥󥹲OutPortΥݥ󥿡
	 * @param inport Subscriber  InPort Υ֥ȥե󥹡
	 * @param id SubscriptionIDˡUUID
	 * @param profile Subscription Υץե롣
	 *
	 * @else
	 *
	 * @param parent OutPort pointer which instansiate Subscriber*.
	 * @param inport An InPort object reference wihch subscribe parent OutPort.
	 * @param id SubscriptionIDUnique UUID.
	 * @param profile Profile of this subscription.
	 *
	 * @endif
	 */
	SubscriberBase(OutPortBase* parent, InPort_ptr inport,
				   const SubscriptionID id, SubscriberProfile profile);

	/*!
	 * @if jp
	 * @brief SubscriberBase 饹ǥȥ饯
	 * @else
	 * @brief SubscriberBase class destructor.
	 * @endif
	 */
	~SubscriberBase();
	
	/*!
	 * @if jp
	 * @brief SubscriberBase ӥ򳫻Ϥ롣
	 * @else
	 * @brief Start SubscriberBase service.
	 * @endif
	 */
	virtual int open(void *args);

	/*!
	 * @if jp
	 * @brief SubscriberBase ӥߤ롣
	 * @else
	 * @brief Stop SubscriberBase service.
	 * @endif
	 */
	virtual int close(unsigned long flags);

	/*!
	 * @if jp
	 * @brief SubscriberBase ӥ (貾۴ؿ)
	 * @else
	 * @brief Main part of SubscriberBase service (pure virtual method).
	 * @endif
	 */
	virtual int svc(void) = 0;
	
	/*!
	 * @if jp
	 * @brief Subscriber ؤΥǡԤ (۴ؿ)
	 * @else
	 * @brief Update data to subscribers. (pure virtual method)
	 * @endif
	 */
	virtual bool update() = 0;
	
	/*!
	 * @if jp
	 * @brief Subscriber ӥλ롣
	 * @else
	 * @brief Stop subscriber service.
	 * @endif
	 */
	virtual void release();

	/*!
	 * @if jp
	 * @brief SubscriberProfile 롣
	 * @else
	 * @brief Get SubscriberProfile.
	 * @endif
	 */
	virtual SubscriberProfile get_profile();

	/*!
	 * @if jp
	 * @brief Subscribe Ƥ InPort Υ֥ȥե󥹤롣
	 * @else
	 * @brief Get InPort's object reference which subscribe this OutPort.
	 * @endif
	 */
	virtual InPort_ptr get_inport();

	/*!
	 * @if jp
	 * @brief SubscSubscribe Ƥ InPort Υ֥ȥե󥹤롣
	 * @else
	 * @brief Get InPort's object reference which subscribe this OutPort.
	 * @endif
	 */
	virtual std::string get_id();

	/*!
	 * @if jp
	 * @brief åɤäƤ뤫ɤȽꤹ롣
	 * @else
	 * @brief Is thread running or not.
	 * @endif
	 */
	inline bool is_running();

	/*!
	 * @if jp
	 * @brief åɤߤ롣ºݤˤϥѿ m_Running  false 
	 * åȤ롣߽ϥ桼ǤʤФʤʤ
	 * @else
	 * @brief Stop this task thread. Actually this method set member variable
	 * "m_Running" to false. User have to implement thread stop logic.
	 * @endif
	 */
	virtual void stop_thread();

  protected:
	//! Request type. once or periodic.
	SubscriberProfile m_Profile;
	
	//! Input port object reference.
	InPort_ptr m_pInPort;
	
	//! Parent outport object reference
	OutPortBase* m_pParent;
	
	//! Subscription ID
	//const SubscriptionID m_Id;
	std::string m_Id;

	//! Thread activity flag (true=running, false=stop)
	bool m_Running;

	struct NewData
	{
	  NewData() : m_Cond(m_Mutex), m_Updated(false) {};
	  bool m_Updated;
#if 0	  
	  ACE_Thread_Mutex m_Mutex;
	  ACE_Condition<ACE_Thread_Mutex> m_Cond;
#else
	  RTC_Thread_Mutex m_Mutex;
	  RTC_Condition<RTC_Thread_Mutex> m_Cond;
#endif
	};

	// A condition variable for data update notification 
	NewData m_NewData;

  };


  //============================================================
  /*!
   * @class SubscriberOnce
   *
   * @brief Subscriber active object who get data once.
   *
   */
  class SubscriberOnce
	: public SubscriberBase 
  {
  public:
	SubscriberOnce(OutPortBase* parent, InPort_ptr inport,
				   const SubscriptionID id, SubscriberProfile profile);
	virtual bool update();
	virtual int svc(void);

  };
  

  /*!
   * @class SubscriberPeriodic
   *
   * @brief Subscriber active object handle subscriber periodic data acquisition
   *
   */
  class SubscriberPeriodic
	: public SubscriberBase
  {
  public:
	SubscriberPeriodic(OutPortBase* parent, InPort_ptr inport,
					   const SubscriptionID id, SubscriberProfile profile);
	virtual bool update();
	virtual int svc(void);
  };


  //============================================================
  /*!
   * @class SubscriberNew
   *
   * @brief Subscriber active object pushes when new data arrived from activity
   *
   */
  class SubscriberNew
	: public SubscriberBase
  {
  public:
	SubscriberNew(OutPortBase* parent, InPort_ptr inport,
				  const SubscriptionID id, SubscriberProfile profile);
	virtual bool update();
	virtual int svc(void);
  };


  //============================================================
  /*!
   * @class SubscriberTriggerd
   *
   * @brief 
   *
   */
  class SubscriberTriggerd
	: public SubscriberBase
  {
  public:
	SubscriberTriggerd(OutPortBase* parent, InPort_ptr inport,
					   const SubscriptionID id, SubscriberProfile profile);
	virtual bool update(){};
	virtual int svc (void){};
  };


  //============================================================
  /*!
   * @class SubscriberPeriodicNew
   *
   * @brief 
   *
   */
  class SubscriberPeriodicNew
	: public SubscriberBase
  {
  public:
	SubscriberPeriodicNew(OutPortBase* parent, InPort_ptr inport,
						  const SubscriptionID id, SubscriberProfile profile);
	virtual bool update(){};
	virtual int svc (void){};
  };


  //============================================================
  /*!
   * @class SubscriberNewPeriodic
   *
   * @brief 
   *
   */
  class SubscriberNewPeriodic
	: public SubscriberBase
  {
  public:
	SubscriberNewPeriodic(OutPortBase* parent, InPort_ptr inport,
						  const SubscriptionID id, SubscriberProfile profile);
	virtual bool update(){};
	virtual int svc (void){};
  };



  //============================================================
  /*!
   * @class SubscriberPeriodicTriggerd
   *
   * @brief 
   *
   */
  class SubscriberPeriodicTriggerd
	: public SubscriberBase
  {
  public:
	SubscriberPeriodicTriggerd(OutPortBase* parent, InPort_ptr inport,
							   const SubscriptionID id, SubscriberProfile profile);
	virtual bool update(){};
	virtual int svc (void){};
  };



  //============================================================
  /*!
   * @class SubscriberTriggerdPeriodic
   *
   * @brief 
   *
   */
  class SubscriberTriggerdPeriodic
	: public SubscriberBase
  {
  public:
	SubscriberTriggerdPeriodic(OutPortBase* parent, InPort_ptr inport,
							   const SubscriptionID id, SubscriberProfile profile);
	virtual bool update(){};
	virtual int svc (void){};
  };





}; // end of namespace RTM



#endif // end of __RtcSubscriber_h__
